@@ -154,24 +154,29 @@ static int	  RemoteHostname;	  /* Log remote hostname from the message */
 static int	  KernLog = 1;		  /* Track kernel logs by default */
 static int	  KeepKernFac;		  /* Keep remotely logged kernel facility */
 static int	  KeepKernTime;		  /* Keep kernel timestamp, evern after initial read */
+static int	  KeepKernConsole;	  /* Keep kernel logging to console */
 
+static int	  rotate_opt;	          /* Set if command line option has been given (wins) */
 static off_t	  RotateSz = 0;		  /* Max file size (bytes) before rotating, disabled by default */
 static int	  RotateCnt = 5;	  /* Max number (count) of log files to keep, set with -c <NUM> */
 
+struct timeval   *retry = NULL;		  /* Set by init() to &init_tv whenever retry jobs exist */
+struct timeval    init_tv = { 5, 0 };	  /* Retry every 5 seconds. */
+
 /*
  * List of notifiers
  */
-static SIMPLEQ_HEAD(notifiers, notifier) nothead = SIMPLEQ_HEAD_INITIALIZER(nothead);
+static TAILQ_HEAD(notifiers, notifier) nothead = TAILQ_HEAD_INITIALIZER(nothead);
 
 /*
  * List of peers and sockets for binding.
  */
-static SIMPLEQ_HEAD(, peer) pqueue = SIMPLEQ_HEAD_INITIALIZER(pqueue);
+static TAILQ_HEAD(peers, peer) pqueue = TAILQ_HEAD_INITIALIZER(pqueue);
 
 /*
  * List fo peers allowed to log to us.
  */
-static SIMPLEQ_HEAD(, allowedpeer) aphead = SIMPLEQ_HEAD_INITIALIZER(aphead);
+static SIMPLEQ_HEAD(allowed, allowedpeer) aphead = SIMPLEQ_HEAD_INITIALIZER(aphead);
 
 /*
  * central list of recognized configuration keywords with an optional
@@ -179,14 +184,8 @@ static SIMPLEQ_HEAD(, allowedpeer) aphead = SIMPLEQ_HEAD_INITIALIZER(aphead);
  * parser moves the argument to the beginning of the parsed line.
  */
 char *secure_str;			  /* string value of secure_mode */
-
-const struct cfkey {
-	const char  *key;
-	char       **var;
-} cfkey[] = {
-	{ "notify",      NULL        },
-	{ "secure_mode", &secure_str },
-};
+char *rotate_sz_str;			  /* string value of RotateSz    */
+char *rotate_cnt_str;			  /* string value of RotateCnt   */
 
 /* Function prototypes. */
 static int  allowaddr(char *s);
@@ -201,24 +200,26 @@ static void rotate_file(struct filed *f, struct stat *stp_or_null);
 static void rotate_all_files(void);
 static void fprintlog_first(struct filed *f, struct buf_msg *buffer);
 static void fprintlog_successive(struct filed *f, int flags);
-void        endtty();
+void        endtty(int);
 void        wallmsg(struct filed *f, struct iovec *iov, int iovcnt);
-void        reapchild();
+void        reapchild(int);
 const char *cvtaddr(struct sockaddr_storage *f, int len);
 const char *cvthname(struct sockaddr *f, socklen_t len);
 static void forw_lookup(struct filed *f);
 void        domark(void *arg);
 void        doflush(void *arg);
-void        debug_switch();
+void        debug_switch(int);
 void        die(int sig);
 static void signal_init(void);
 static void boot_time_init(void);
+static void retry_init(void);
 static void init(void);
 static int  strtobytes(char *arg);
-static int  cfparse(FILE *fp, struct files *newf, struct notifiers *newn);
+static void cflisten(char *ptr, void *arg);
+static int  cfparse(FILE *fp, struct files *newf);
 int         decode(char *name, struct _code *codetab);
 static void logit(char *, ...);
-static void notifier_add(struct notifiers *newn, const char *program);
+static void notifier_add(char *program, void *arg);
 static void notifier_invoke(const char *logfile);
 static void notifier_free_all(void);
 void        reload(int);
@@ -227,6 +228,21 @@ static int  validate(struct sockaddr *sa, const char *hname);
 static int  waitdaemon(int);
 static void timedout(int);
 
+/*
+ * Configuration file keywords, variables, and optional callbacks
+ */
+const struct cfkey {
+	const char  *key;
+	char       **var;
+	void       (*cb)(char *, void *);
+	void        *arg;
+} cfkey[] = {
+	{ "listen",       NULL,            cflisten, NULL         },
+	{ "notify",       NULL,            notifier_add, &nothead },
+	{ "rotate_size",  &rotate_sz_str,  NULL, NULL             },
+	{ "rotate_count", &rotate_cnt_str, NULL, NULL             },
+	{ "secure_mode",  &secure_str,     NULL, NULL             },
+};
 
 /*
  * Very basic, and incomplete, check if we're running in a container.
@@ -266,11 +282,24 @@ static int addpeer(struct peer *pe0)
 {
 	struct peer *pe;
 
+	TAILQ_FOREACH(pe, &pqueue, pe_link) {
+		if (((pe->pe_name == NULL && pe0->pe_name == NULL) ||
+		     (pe->pe_name != NULL && pe0->pe_name != NULL && strcmp(pe->pe_name, pe0->pe_name) == 0)) &&
+		    ((pe->pe_serv == NULL && pe0->pe_serv == NULL) ||
+		     (pe->pe_serv != NULL && pe0->pe_serv != NULL && strcmp(pe->pe_serv, pe0->pe_serv) == 0))) {
+			/* update flags */
+			pe->pe_mark = pe0->pe_mark;
+			pe->pe_mode = pe0->pe_mode;
+
+			return 0;
+		}
+	}
+
 	pe = calloc(1, sizeof(*pe));
 	if (pe == NULL)
 		err(1, "malloc failed");
 	*pe = *pe0;
-	SIMPLEQ_INSERT_TAIL(&pqueue, pe, pe_link);
+	TAILQ_INSERT_TAIL(&pqueue, pe, pe_link);
 
 	return 0;
 }
@@ -363,6 +392,9 @@ int usage(int code)
 	       "  -H        Use hostname from message instead of address for remote messages\n"
 	       "  -K        Disable kernel logging, useful in container use-cases\n"
 	       "  -k        Allow logging with facility 'kernel', otherwise remapped to 'user'\n"
+#ifdef __linux__
+	       "  -l        Keep kernel logging to console, use sysctl to adjust kernel.printk\n"
+#endif
 	       "  -m MINS   Interval between MARK messages, 0 to disable, default: 20 min\n"
 	       "  -n        Disable DNS query for every request\n"
 	       "  -P FILE   File to store the process ID, default: %s\n"
@@ -393,11 +425,10 @@ int main(int argc, char *argv[])
 	pid_t ppid = 1;
 	int no_sys = 0;
 	int pflag = 0;
-	int bflag = 0;
 	char *ptr;
 	int ch;
 
-	while ((ch = getopt(argc, argv, "468Aa:b:C:cdHFf:Kkm:nP:p:r:sTtv?")) != EOF) {
+	while ((ch = getopt(argc, argv, "468Aa:b:C:cdHFf:Kklm:nP:p:r:sTtv?")) != EOF) {
 		switch ((char)ch) {
 		case '4':
 			family = PF_INET;
@@ -421,13 +452,13 @@ int main(int argc, char *argv[])
 			break;
 
 		case 'b':
-			bflag = 1;
 			ptr = strchr(optarg, ':');
 			if (ptr)
 				*ptr++ = 0;
 			addpeer(&(struct peer) {
 				.pe_name = optarg,
 				.pe_serv = ptr,
+				.pe_mark = -1,
 			});
 			break;
 
@@ -464,6 +495,10 @@ int main(int argc, char *argv[])
 			KeepKernFac = 1;
 			break;
 
+		case 'l':
+			KeepKernConsole = 1;
+			break;
+
 		case 'm': /* mark interval */
 			MarkInterval = atoi(optarg) * 60;
 			break;
@@ -486,11 +521,13 @@ int main(int argc, char *argv[])
 			addpeer(&(struct peer) {
 				.pe_name = optarg,
 				.pe_mode = 0666,
+				.pe_mark = -1,
 			});
 			break;
 
 		case 'r':
 			parse_rotation(optarg, &RotateSz, &RotateCnt);
+			rotate_opt++;
 			break;
 
 		case 's':
@@ -520,13 +557,6 @@ int main(int argc, char *argv[])
 	if ((argc -= optind))
 		return usage(1);
 
-	/* Default to listen to :514 (syslog/udp) */
-	if (!bflag)
-		addpeer(&(struct peer) {
-				.pe_name = NULL,
-				.pe_serv = "syslog",
-			});
-
 	/* Figure out where to read system log messages from */
 	if (!pflag) {
 		/* Do we run under systemd-journald (Requires=syslog.socket)? */
@@ -538,6 +568,7 @@ int main(int argc, char *argv[])
 			addpeer(&(struct peer) {
 					.pe_name = _PATH_LOG,
 					.pe_mode = 0666,
+					.pe_mark = -1,
 				});
 		}
 	}
@@ -615,7 +646,7 @@ no_klogd:
 	for (;;) {
 		int rc;
 
-		rc = socket_poll(NULL);
+		rc = socket_poll(retry);
 		if (restart > 0) {
 			restart--;
 			logit("\nReceived SIGHUP, reloading syslogd.\n");
@@ -633,8 +664,11 @@ no_klogd:
 			rotate_all_files();
 		}
 
-		if (rc < 0 && errno != EINTR)
-			ERR("select()");
+		if (rc < 0) {
+			if (errno != EINTR)
+				ERR("select()");
+		} else if (rc == 0)
+			retry_init();
 
 		if (KernLog)
 			sys_seqno_save();
@@ -742,14 +776,14 @@ static void unix_cb(int sd, void *arg)
 	parsemsg(LocalHostName, msg);
 }
 
-static void create_unix_socket(struct peer *pe)
+static int create_unix_socket(struct peer *pe)
 {
 	struct sockaddr_un sun;
 	struct addrinfo ai;
 	int sd = -1;
 
 	if (pe->pe_socknum)
-		return;		/* Already set up */
+		return 0;	/* Already set up */
 
 	memset(&ai, 0, sizeof(ai));
 	ai.ai_addr = (struct sockaddr *)&sun;
@@ -763,11 +797,12 @@ static void create_unix_socket(struct peer *pe)
 	if (sd < 0)
 		goto err;
 
-	logit("Created UNIX socket %d ...\n", sd);
+	NOTE("Created UNIX socket %s", sun.sun_path);
 	pe->pe_sock[pe->pe_socknum++] = sd;
-	return;
+	return 0;
 err:
 	ERR("cannot create %s", pe->pe_name);
+	return 1;
 }
 
 static void unmapped(struct sockaddr *sa)
@@ -823,6 +858,11 @@ static void inet_cb(int sd, void *arg)
 	parsemsg(hname, msg);
 }
 
+/*
+ * Depending on the setup of /etc/resolv.conf, and the system resolver,
+ * a call to this function may be blocked for 10 seconds, or even more,
+ * waiting for a response.  See https://serverfault.com/a/562108/122484
+ */
 static int nslookup(const char *host, const char *service, struct addrinfo **ai)
 {
 	struct addrinfo hints;
@@ -834,7 +874,7 @@ static int nslookup(const char *host, const char *service, struct addrinfo **ai)
 	/* Reset resolver cache and retry name lookup */
 	res_init();
 
-	logit("nslookup '%s:%s'\n", node ?: "none", service);
+	logit("nslookup '%s:%s'\n", node ?: "*", service);
 	memset(&hints, 0, sizeof(hints));
 	hints.ai_flags    = !node ? AI_PASSIVE : 0;
 	hints.ai_family   = family;
@@ -843,22 +883,25 @@ static int nslookup(const char *host, const char *service, struct addrinfo **ai)
 	return getaddrinfo(node, service, &hints, ai);
 }
 
-static void create_inet_socket(struct peer *pe)
+static int create_inet_socket(struct peer *pe)
 {
 	struct addrinfo *ai, *res;
-	int sd, err;
+	int err, rc = 0;
+
+	if (pe->pe_socknum)
+		return 0;	/* Already set up */
 
 	err = nslookup(pe->pe_name, pe->pe_serv, &res);
 	if (err) {
-		ERRX("%s/udp service unknown: %s", pe->pe_serv,
-		     gai_strerror(err));
-		return;
+		ERRX("%s:%s/udp service unknown: %s", pe->pe_name ?: "*", pe->pe_serv, gai_strerror(err));
+		return 1;
 	}
 
 	for (ai = res; ai; ai = ai->ai_next) {
+		int sd;
+
 		if (pe->pe_socknum + 1 >= NELEMS(pe->pe_sock)) {
-			WARN("Only %zd IP addresses per socket supported.",
-			     NELEMS(pe->pe_sock));
+			WARN("Only %zd IP addresses per socket supported.", NELEMS(pe->pe_sock));
 			break;
 		}
 
@@ -868,15 +911,35 @@ static void create_inet_socket(struct peer *pe)
 			ai->ai_flags &= ~AI_SECURE;
 
 		sd = socket_create(ai, inet_cb, NULL);
-		if (sd < 0)
+		if (sd < 0) {
+			WARN("Failed creating socket for %s:%s: %s", pe->pe_name ?: "*",
+			      pe->pe_serv, strerror(errno));
+			rc = 1;
 			continue;
+		}
 
-		logit("Created inet socket %d for %s:%s ...\n", sd,
-		      pe->pe_name, pe->pe_serv);
+		if (!SecureMode) {
+			pe->pe_mode |= 01000;
+			NOTE("Opened inet socket %s:%s", pe->pe_name ?: "*", pe->pe_serv);
+		}
 		pe->pe_sock[pe->pe_socknum++] = sd;
 	}
 
 	freeaddrinfo(res);
+	if (rc && pe->pe_socknum == 0)
+		return rc;
+
+	return 0;
+}
+
+static void close_socket(struct peer *pe)
+{
+	for (size_t i = 0; i < pe->pe_socknum; i++) {
+		if (pe->pe_mode & 01000)
+			NOTE("Closing inet socket %s:%s", pe->pe_name ?: "*", pe->pe_serv);
+		socket_close(pe->pe_sock[i]);
+	}
+	pe->pe_socknum = 0;
 }
 
 void untty(void)
@@ -1718,21 +1781,34 @@ static void logmsg(struct buf_msg *buffer)
 static void logrotate(struct filed *f)
 {
 	struct stat statf;
+	off_t sz;
 
-	if (!f->f_rotatesz)
+	if (!f->f_rotatesz && !RotateSz)
 		return;
 
+	if (f->f_rotatesz)
+		sz = f->f_rotatesz;
+	else
+		sz = RotateSz;
+
 	if (fstat(f->f_file, &statf))
 		return;
 
 	/* bug (mostly harmless): can wrap around if file > 4gb */
-	if (S_ISREG(statf.st_mode) && statf.st_size > f->f_rotatesz)
+	if (S_ISREG(statf.st_mode) && statf.st_size > sz)
 		rotate_file(f, &statf);
 }
 
 static void rotate_file(struct filed *f, struct stat *stp_or_null)
 {
-	if (f->f_rotatecount > 0) { /* always 0..999 */
+	int cnt;
+
+	if (f->f_rotatecount)
+		cnt = f->f_rotatecount;
+	else
+		cnt = RotateCnt;
+
+	if (cnt > 0) { /* always 0..999 */
 		struct stat st_stack;
 		int  len = strlen(f->f_un.f_fname) + 10 + 5;
 		int  i;
@@ -1740,7 +1816,7 @@ static void rotate_file(struct filed *f, struct stat *stp_or_null)
 		char newFile[len];
 
 		/* First age zipped log files */
-		for (i = f->f_rotatecount; i > 1; i--) {
+		for (i = cnt; i > 1; i--) {
 			snprintf(oldFile, len, "%s.%d.gz", f->f_un.f_fname, i - 1);
 			snprintf(newFile, len, "%s.%d.gz", f->f_un.f_fname, i);
 
@@ -1754,10 +1830,11 @@ static void rotate_file(struct filed *f, struct stat *stp_or_null)
 			snprintf(newFile, len, "%s.%d", f->f_un.f_fname, i);
 
 			if (!rename(oldFile, newFile) && i > 0) {
-				size_t clen = 18 + strlen(newFile) + 1;
+				const char *gzip = "gzip -f";
+				size_t clen = strlen(gzip) + len + 1;
 				char cmd[clen];
 
-				snprintf(cmd, sizeof(cmd), "gzip -f %s", newFile);
+				snprintf(cmd, clen, "%s %s", gzip, newFile);
 				system(cmd);
 			}
 		}
@@ -1783,7 +1860,7 @@ static void rotate_file(struct filed *f, struct stat *stp_or_null)
 			return;
 		}
 
-		if (!SIMPLEQ_EMPTY(&nothead))
+		if (!TAILQ_EMPTY(&nothead))
 			notifier_invoke(f->f_un.f_fname);
 	}
 	ftruncate(f->f_file, 0);
@@ -1794,7 +1871,14 @@ static void rotate_all_files(void)
 	struct filed *f;
 
 	SIMPLEQ_FOREACH(f, &fhead, f_link) {
-		if (f->f_type == F_FILE && f->f_rotatesz)
+		off_t sz;
+
+		if (f->f_rotatesz)
+			sz = f->f_rotatesz;
+		else
+			sz = RotateSz;
+
+		if (f->f_type == F_FILE && sz)
 			rotate_file(f, NULL);
 	}
 }
@@ -1894,6 +1978,7 @@ void fprintlog_write(struct filed *f, struct iovec *iov, int iovcnt, int flags)
 		if (lsent != len) {
 			switch (errno) {
 			case ENOBUFS:
+			case ENONET:		/* returned by socket_ffs() */
 			case ENETDOWN:
 			case ENETUNREACH:
 			case EHOSTUNREACH:
@@ -1994,7 +2079,9 @@ void fprintlog_write(struct filed *f, struct iovec *iov, int iovcnt, int flags)
 		f->f_time = timer_now();
 		logit("\n");
 		pushiov(iov, iovcnt, "\r\n");
-		wallmsg(f, &iov[0], iovcnt);
+		/* Make sure it's seen, may be followed by reboot */
+		pushiov(iov, iovcnt, "\r\n");
+		wallmsg(f, iov, iovcnt);
 		break;
 	} /* switch */
 
@@ -2371,8 +2458,8 @@ static void forw_lookup(struct filed *f)
 	char *host = f->f_un.f_forw.f_hname;
 	char *serv = f->f_un.f_forw.f_serv;
 	struct addrinfo *ai;
+	time_t now, diff;
 	int err, first;
-	time_t diff;
 
 	if (SecureMode > 1) {
 		if (f->f_un.f_forw.f_addr)
@@ -2385,27 +2472,27 @@ static void forw_lookup(struct filed *f)
 	/* Called from cfline() for initial lookup? */
 	first = f->f_type == F_UNUSED ? 1 : 0;
 
+	now = timer_now();
+	diff = now - f->f_time;
+
 	/*
-	 * Not INET_SUSPEND_TIME, but back off a few seconds at least
-	 * to prevent syslogd from hammering the resolver for every
-	 * little message that is logged.  E.g., at boot when we read
-	 * the kernel ring buffer.
+	 * Back off a minute to prevent unnecessary delays on other log
+	 * targets due to being blockd in nslookup().  This means bootup
+	 * log messages may not be logged for named remote targets since
+	 * networking may not be available until later.
 	 */
-	diff = timer_now() - f->f_time;
-	if (!first && diff < 5)
+	if (!first && diff < INET_DNS_DELAY)
 		return;
 
 	err = nslookup(host, serv, &ai);
 	if (err) {
 		f->f_type = F_FORW_UNKN;
-		f->f_time = timer_now();
-		if (!first && !(f->f_flags & SUSP_RETR))
+		f->f_time = now;
+		if (!first)
 			WARN("Failed resolving '%s:%s': %s", host, serv, gai_strerror(err));
-		f->f_flags |= SUSP_RETR; /* Retry silently */
 		return;
 	}
 
-	f->f_flags &= ~SUSP_RETR;
 	f->f_type = F_FORW;
 	f->f_un.f_forw.f_addr = ai;
 	f->f_prevcount = 0;
@@ -2493,22 +2580,20 @@ void die(int signo)
 	 */
 	timer_exit();
 
+	/*
+	 * Close all UNIX and inet sockets
+	 */
+	TAILQ_FOREACH_SAFE(pe, &pqueue, pe_link, next) {
+		TAILQ_REMOVE(&pqueue, pe, pe_link);
+		close_socket(pe);
+		free(pe);
+	}
+
 	/*
 	 * Close all open log files.
 	 */
 	close_open_log_files();
 
-	/*
-	 * Close all UNIX and inet sockets
-	 */
-	SIMPLEQ_FOREACH_SAFE(pe, &pqueue, pe_link, next) {
-		for (size_t i = 0; i < pe->pe_socknum; i++) {
-			logit("Closing socket %d ...\n", pe->pe_sock[i]);
-			socket_close(pe->pe_sock[i]);
-		}
-		free(pe);
-	}
-
 	kern_console_on();
 
 	exit(0);
@@ -2663,15 +2748,40 @@ static void boot_time_init(void)
 #endif
 }
 
+/*
+ * Used by init() to trigger retries of, e.g., binding to interfaces.
+ */
+static void retry_init(void)
+{
+	struct peer *pe;
+	int fail = 0;
+
+	logit("Retrying socket init ...\n");
+	TAILQ_FOREACH(pe, &pqueue, pe_link) {
+		if (pe->pe_name && pe->pe_name[0] == '/') {
+			fail |= create_unix_socket(pe);
+		} else {
+			/* skip any marked for deletion */
+			if (SecureMode < 2)
+				fail |= create_inet_socket(pe);
+		}
+	}
+
+	if (!fail) {
+		logit("Socket re-init done.\n");
+		retry = NULL;
+	}
+}
+
 /*
  *  INIT -- Initialize syslogd from configuration table
  */
 static void init(void)
 {
-	struct notifiers newn = SIMPLEQ_HEAD_INITIALIZER(newn);
 	struct files newf = SIMPLEQ_HEAD_INITIALIZER(newf);
+	struct peer *pe, *penext;
+	int bflag = 0, fail = 0;
 	struct filed *f;
-	struct peer *pe;
 	FILE *fp;
 	char *p;
 
@@ -2697,7 +2807,7 @@ static void init(void)
 		 * Good software also always checks its return values...
 		 * If syslogd starts up before DNS is up & /etc/hosts
 		 * doesn't have LocalHostName listed, gethostbyname will
-		 * return NULL. 
+		 * return NULL.
 		 */
 		hent = gethostbyname(LocalHostName);
 		if (hent)
@@ -2722,6 +2832,21 @@ static void init(void)
 	 */
 	tzset();
 
+	/*
+	 * mark
+	 */
+	TAILQ_FOREACH(pe, &pqueue, pe_link) {
+		if (pe->pe_mark == -1)
+			continue;
+
+		pe->pe_mark = 1;
+	}
+
+	/*
+	 * Free all notifiers
+	 */
+	notifier_free_all();
+
 	/*
 	 * Read configuration file(s)
 	 */
@@ -2736,7 +2861,7 @@ static void init(void)
 		}
 	}
 
-	if (cfparse(fp, &newf, &newn)) {
+	if (cfparse(fp, &newf)) {
 		fclose(fp);
 		return;
 	}
@@ -2750,36 +2875,70 @@ static void init(void)
 	fhead = newf;
 
 	/*
-	 * Free all notifiers
+	 * Ensure a default listen *:514 exists (compat)
 	 */
-	notifier_free_all();
-
-	nothead = newn;
+	TAILQ_FOREACH(pe, &pqueue, pe_link) {
+		if (pe->pe_mark == 1)
+			continue; /* marked for deletion */
+		if (pe->pe_name && pe->pe_name[0] == '/')
+			continue; /* named pipe */
+		if (pe->pe_name || pe->pe_serv) {
+			bflag = 1;
+			break;	/* static or from .conf */
+		}
+	}
+	if (!bflag) {
+		/* Default to listen to :514 (syslog/udp) */
+		addpeer(&(struct peer) {
+				.pe_name = NULL,
+				.pe_serv = "514",
+			});
+	}
 
 	/*
-	 * Open or close sockets for local and remote communication
+	 * Sweep
 	 */
-	SIMPLEQ_FOREACH(pe, &pqueue, pe_link) {
-		if (pe->pe_name && pe->pe_name[0] == '/') {
-			create_unix_socket(pe);
-		} else {
-			for (size_t i = 0; i < pe->pe_socknum; i++)
-				socket_close(pe->pe_sock[i]);
-			pe->pe_socknum = 0;
+	TAILQ_FOREACH_SAFE(pe, &pqueue, pe_link, penext) {
+		if (pe->pe_mark <= 0)
+			continue;
 
-			if (SecureMode < 2)
-				create_inet_socket(pe);
-		}
+		TAILQ_REMOVE(&pqueue, pe, pe_link);
+		close_socket(pe);
+		free(pe);
 	}
 
 	Initialized = 1;
 
+	flog(LOG_SYSLOG | LOG_INFO, "syslogd v" VERSION ": restart.");
+	logit("syslogd: restarted.\n");
+
+	/*
+	 * Open or close sockets for local and remote communication
+	 * These may be delayed, so start local logging first.
+	 */
+	TAILQ_FOREACH(pe, &pqueue, pe_link) {
+		if (pe->pe_name && pe->pe_name[0] == '/') {
+			fail |= create_unix_socket(pe);
+		} else {
+			close_socket(pe);
+
+			if (SecureMode < 2)
+				fail |= create_inet_socket(pe);
+		}
+	}
+
+	if (fail)
+		retry = &init_tv;
+	else
+		retry = NULL;
+
 	if (Debug) {
-		if (!SIMPLEQ_EMPTY(&nothead)) {
+		if (!TAILQ_EMPTY(&nothead)) {
 			struct notifier *np;
 
-			SIMPLEQ_FOREACH(np, &nothead, n_link)
+			TAILQ_FOREACH(np, &nothead, n_link) {
 				printf("notify %s\n", np->n_program);
+			}
 			printf("\n");
 		}
 
@@ -2827,9 +2986,19 @@ static void init(void)
 			printf("\n");
 		}
 	}
+}
 
-	flog(LOG_SYSLOG | LOG_INFO, "syslogd v" VERSION ": restart.");
-	logit("syslogd: restarted.\n");
+static void cflisten(char *ptr, void *arg)
+{
+	char *peer = ptr;
+
+	ptr = strchr(peer, ':');
+	if (ptr)
+		*ptr++ = 0;
+	addpeer(&(struct peer) {
+			.pe_name = peer,
+			.pe_serv = ptr,
+		});
 }
 
 static void cfrot(char *ptr, struct filed *f)
@@ -2842,13 +3011,12 @@ static void cfrot(char *ptr, struct filed *f)
 		*c++ = 0;
 		cnt = atoi(c);
 	}
+	if (cnt > 0)
+		f->f_rotatecount = cnt;
 
 	sz = strtobytes(ptr);
-	if (sz > 0 && cnt > 0) {
-		logit("Set rotate size %d bytes, %d rotations\n", sz, cnt);
-		f->f_rotatecount = cnt;
+	if (sz > 0)
 		f->f_rotatesz = sz;
-	}
 }
 
 static int cfopt(char **ptr, const char *opt)
@@ -3055,7 +3223,7 @@ static struct filed *cfline(char *line)
 		if (bp)
 			*bp++ = 0;
 		else
-			bp = "syslog"; /* default: 514/udp */
+			bp = "514"; /* default: 514/udp */
 
 		strlcpy(f->f_un.f_forw.f_hname, p, sizeof(f->f_un.f_forw.f_hname));
 		strlcpy(f->f_un.f_forw.f_serv, bp, sizeof(f->f_un.f_forw.f_serv));
@@ -3126,13 +3294,6 @@ static struct filed *cfline(char *line)
 		break;
 
 	case F_FILE:
-		/* default rotate from command line */
-		if (f->f_rotatesz == 0) {
-			f->f_rotatecount = RotateCnt;
-			f->f_rotatesz = RotateSz;
-		}
-		/* fallthrough */
-
 	default:
 		/* All other targets default to RFC3164 */
 		if (f->f_flags & (RFC3164 | RFC5424))
@@ -3176,6 +3337,8 @@ const struct cfkey *cfkey_match(char *cline)
 
 		if (cfk->var)
 			*cfk->var = strdup(p);
+		else if (cfk->cb)
+			cfk->cb(p, cfk->arg);
 		else
 			memmove(cline, p, strlen(p) + 1);
 
@@ -3188,7 +3351,7 @@ const struct cfkey *cfkey_match(char *cline)
 /*
  * Parse .conf file and append to list
  */
-static int cfparse(FILE *fp, struct files *newf, struct notifiers *newn)
+static int cfparse(FILE *fp, struct files *newf)
 {
 	const struct cfkey *cfk;
 	struct filed *f;
@@ -3239,7 +3402,7 @@ static int cfparse(FILE *fp, struct files *newf, struct notifiers *newn)
 				p++;
 
 			logit("Searching for %s ...\n", p);
-			if (glob(p, GLOB_NOSORT, NULL, &gl))
+			if (glob(p, 0, NULL, &gl))
 				logit("No files match %s\n", p);
 
 			for (size_t i = 0; i < gl.gl_pathc; i++) {
@@ -3254,7 +3417,7 @@ static int cfparse(FILE *fp, struct files *newf, struct notifiers *newn)
 				}
 
 				logit("Parsing %s ...\n", gl.gl_pathv[i]);
-				cfparse(fpi, newf, newn);
+				cfparse(fpi, newf);
 				fclose(fpi);
 			}
 			globfree(&gl);
@@ -3262,11 +3425,8 @@ static int cfparse(FILE *fp, struct files *newf, struct notifiers *newn)
 		}
 
 		cfk = cfkey_match(cline);
-		if (cfk) {
-			if (!strcmp(cfk->key, "notify"))
-				notifier_add(newn, cline);
+		if (cfk)
 			continue;
-		}
 
 		f = cfline(cline);
 		if (!f)
@@ -3288,6 +3448,32 @@ static int cfparse(FILE *fp, struct files *newf, struct notifiers *newn)
 		secure_str = NULL;
 	}
 
+	if (rotate_sz_str) {
+		if (rotate_opt) {
+			logit("Skipping 'rotate_size', already set on command line.");
+		} else {
+			int val = strtobytes(rotate_sz_str);
+			if (val > 0)
+				RotateSz = val;
+		}
+
+		free(rotate_sz_str);
+		rotate_sz_str = NULL;
+	}
+
+	if (rotate_cnt_str) {
+		if (rotate_opt) {
+			logit("Skipping 'rotate_count', already set on command line.");
+		} else {
+			int val = atoi(rotate_cnt_str);
+			if (val > 0)
+				RotateCnt = val;
+		}
+
+		free(rotate_cnt_str);
+		rotate_cnt_str = NULL;
+	}
+
 	return 0;
 }
 
@@ -3367,7 +3553,7 @@ static int allowaddr(char *s)
 				goto err;
 		}
 	} else {
-		if ((se = getservbyname("syslog", "udp")))
+		if ((se = getservbyname("514", "udp")))
 			ap->port = ntohs(se->s_port);
 		else
 			/* sanity, should not happen */
@@ -3595,8 +3781,7 @@ static int validate(struct sockaddr *sa, const char *hname)
 			} else
 				continue;
 		} else {
-			if (fnmatch(ap->a_name, name, FNM_NOESCAPE) ==
-			    FNM_NOMATCH) {
+			if (fnmatch(ap->a_name, name, FNM_NOESCAPE) == FNM_NOMATCH) {
 				logit("rejected in rule %d due to name mismatch.\n", i);
 				continue;
 			}
@@ -3623,8 +3808,10 @@ static void logit(char *fmt, ...)
 	fflush(stdout);
 }
 
-static void notifier_add(struct notifiers *newn, const char *program)
+static void notifier_add(char *program, void *arg)
 {
+	struct notifiers *newn = (struct notifiers *)arg;
+
 	while (*program && isspace(*program))
 		++program;
 
@@ -3643,7 +3830,7 @@ static void notifier_add(struct notifiers *newn, const char *program)
 			ERR("Cannot allocate memory for a notify program");
 			return;
 		}
-		SIMPLEQ_INSERT_TAIL(newn, np, n_link);
+		TAILQ_INSERT_TAIL(newn, np, n_link);
 	} else
 		logit("notify: non-existing, or not executable program\n");
 }
@@ -3656,7 +3843,7 @@ static void notifier_invoke(const char *logfile)
 
 	logit("notify: rotated %s, invoking hooks\n", logfile);
 
-	SIMPLEQ_FOREACH(np, &nothead, n_link) {
+	TAILQ_FOREACH(np, &nothead, n_link) {
 		childpid = fork();
 
 		switch (childpid) {
@@ -3681,7 +3868,8 @@ static void notifier_free_all(void)
 {
 	struct notifier *np, *npnext;
 
-	SIMPLEQ_FOREACH_SAFE(np, &nothead, n_link, npnext) {
+	TAILQ_FOREACH_SAFE(np, &nothead, n_link, npnext) {
+		TAILQ_REMOVE(&nothead, np, n_link);
 		free(np->n_program);
 		free(np);
 	}
